#!/usr/bin/env python
""" Periodic Task
# Author: irox_rl
# Purpose: Periodic Task to run tasks as required per application
# Version 1.0.3
"""

# local imports #
from .channels import clear_channel_messages

# non-local imports
import datetime
import discord
from discord.ext import tasks
from typing import Callable


class PeriodicTask:
    """ Periodic Task
            This class uses ***@task***s to run tasks as required.
            """

    def __init__(self,
                 cycle_time: int,
                 master_bot,
                 task_callback: Callable = None,
                 enable_admin: bool = False):
        """ Initialize method\n
                            **param cycle_time**: time interval (in seconds) for the periodic task to run\n
                            **param master_bot**: reference of the Bot this task belongs to; for callbacks\n
                            **param enable_admin**: enable_admin functions to be performed by this bot\n
                            **param admin_channel**: admin channel that this bot can post admin info to\n
                            **param roster_channel**: roster channel that this bot can post roster info to\n
                            All data is initialized to zero. Franchise load will be called 'on_ready' of the bot\n
                            **returns**: None\n
                        """
        self.callback: Callable = task_callback
        self.admin_message: discord.Message | None = None
        self.bot = master_bot
        self.cycle_time: int = cycle_time
        self._enable_admin: bool = enable_admin
        self._initialized: bool = False
        self.on_tick: [Callable] = []
        self.ticks: int = 0

    @property
    def initialized(self):
        return self._initialized

    @tasks.loop()
    async def run(self):
        """ looping task\n
            **returns**: None\n
            """
        await self.admin()
        await self.callback()

    async def __regen_admin__(self):
        """ helper function to clear out and repost admin information\n
        **returns**: None\n
        """
        """ if there is no admin channel, do not run
        """
        if not self.bot.admin_channel or not self._enable_admin:
            return

        """ clear out channel messages
        """
        await clear_channel_messages(self.bot.admin_channel, 100)

        """ generate new admin message and record it
        """
        self.admin_message = await self.bot.admin_channel.send(embed=self.bot.info_embed())

    def __time__(self):
        """ periodic task time function\n
                    **returns**: None\n
                    """
        self.bot._last_time = self.bot._time if self.bot._time else datetime.datetime.now()
        self.bot._time = datetime.datetime.now()
        self.bot.new_weekday = True if self.bot._time.weekday() != self.bot._last_time.weekday() else False
        self.bot.new_week = True if self.bot._time.isocalendar().week != self.bot._last_time.isocalendar().week else False

    async def admin(self):
        """ periodic task admin function\n
            **returns**: None\n
            """
        self.ticks += 1
        self.__time__()

        for callback in self.on_tick:
            callback(self.ticks)

        self._initialized = True

        if self._enable_admin and self.bot.admin_channel:
            try:
                await self.admin_message.edit(embed=self.bot.info_embed())
            except discord.errors.NotFound:
                await self.__regen_admin__()
            except AttributeError:
                await self.__regen_admin__()
            except discord.errors.DiscordServerError:
                await self.__regen_admin__()

    def change_interval(self,
                        seconds: int) -> None:
        """ change interval of periodic task\n
        **param seconds**: period that the task will sleep between intervals\n
        **returns**: self\n
        """
        self.cycle_time = seconds
        self.run.change_interval(seconds=self.cycle_time)



#!/usr/bin/env python
""" Minor League E-Sports Bot
# Author: irox_rl
# Purpose: General Functions of a League Franchise summarized in bot fashion!
# Version 1.0.5
#
# v1.0.5 - update bot to allow list command prefixes, as well as string
            update on_command_error to include multiple error types
            remove unused method
            include 'last_time' to public properties
            include 'loaded' property for all external tasks.
            updated embed formatting
"""

# local imports #
from .channels import get_channel_by_id
from .commands import Commands
from .err import register_callback, InsufficientPrivilege, IllegalChannel, BotNotLoaded
from .periodic_task import PeriodicTask

# non-local imports
import asyncio
import datetime
import discord
from discord.ext import commands as disco_commands
import dotenv
import os


class Bot(discord.ext.commands.Bot):
    """ Default bot by irox
    """

    def __init__(self,
                 command_prefix: str | list,
                 bot_intents: discord.Intents | None,
                 command_cogs: [disco_commands.Cog]):
        super().__init__(command_prefix=command_prefix,
                         intents=bot_intents,
                         help_command=None,
                         case_insensitive=True)
        self._version = os.getenv('VERSION')
        if not self._version:
            raise ValueError('Version not determined from .env file. Please use a valid .env file.')
        try:
            self._server_icon = int(os.getenv('SERVER_ICON'))
        except ValueError:
            self._server_icon = None

        self._handler: discord.User | None = None
        self._admin_channel: discord.TextChannel | None = None
        self._notification_channel: discord.TextChannel | None = None
        self._admin_commands_channel: discord.TextChannel | None = None
        self._public_commands_channel: discord.TextChannel | None = None
        self._extended_public_commands_channels: [discord.TextChannel] = []
        self._initialized = False
        self._guild: discord.Guild | None = None
        self._time = datetime.datetime.now()
        self._last_time = self._time
        self._start_time = datetime.datetime.now()
        self._cycle_time = int(os.getenv('CYCLE_TIME'))
        self._periodic_task: PeriodicTask = PeriodicTask(self.cycle_time,
                                                         self,
                                                         task_callback=self.on_task,
                                                         enable_admin=True)
        for cog in command_cogs:
            asyncio.run(self.add_cog(cog(self)))

        register_callback(self.__err__)

    @property
    def admin_channel(self) -> discord.TextChannel | None:
        """ return the admin channel
        """
        return self._admin_channel

    @property
    def admin_commands_channel(self) -> discord.TextChannel | None:
        return self._admin_commands_channel

    @property
    def cycle_time(self) -> int:
        """ return the cycle time
                """
        return self._cycle_time

    @property
    def day_of_week(self) -> str:
        """ return the day of week, formatted as yyyy-mm-dd hh:mm:ss
                """
        return self._time.strftime('%A')

    @property
    def default_embed_color(self):
        """ return the discord.color for an embed object to use
            Over-ride this property to use a different color
        """
        return discord.Color.dark_red()

    @property
    def guild(self) -> discord.Guild:
        return self._guild

    @property
    def handler(self) -> discord.User | None:
        """ return the handler of this bot
                        """
        return self._handler

    @property
    def loaded(self) -> bool:
        """ loaded status for bot\n
        override to include external processes that need to load when initiating
        """
        return self._periodic_task.initialized and self._initialized

    @property
    def notification_channel(self) -> discord.TextChannel | None:
        return self._notification_channel

    @property
    def public_commands_channel(self) -> discord.TextChannel | None:
        return self._public_commands_channel

    @property
    def extended_public_commands_channels(self) -> [discord.TextChannel]:
        return self._extended_public_commands_channels

    @extended_public_commands_channels.setter
    def extended_public_commands_channels(self,
                                          value):
        self._extended_public_commands_channels = value

    @property
    def last_time(self):
        return self._last_time

    @property
    def server_icon(self):
        return self._server_icon

    @property
    def time(self) -> datetime.datetime:
        return self._time

    @property
    def version(self) -> str:
        """ return the version of this repo
                                """
        return self._version

    async def __begin_task__(self) -> None:
        """ Begin periodic task\n
        **returns**: None\n
        """
        self._periodic_task.change_interval(seconds=self.cycle_time)
        self._periodic_task.run.start()

    def default_embed(self,
                      title: str,
                      description: str = '') -> discord.Embed:
        """ Helper function to easily and repeatedly get the same embed\n
            **param title**: title of the embed\n
            **param description**: description to create the embed with\n
            **returns**: discord.Embed with name and description supplied
         """
        # this should be updated to be a class method, along with the default color
        return discord.Embed(color=self.default_embed_color,
                             title=title,
                             description=description)

    async def __err__(self,
                      message: str | Exception) -> None:
        """ Helper function to send error or notification messages to notify channel with a single parameter.\n
            **If a notification channel does not exist**, the notification is printed to console instead\n
            **param message**: message to report\n
            **returns**: None\n
            """
        if not message:
            return
        if not self._notification_channel:
            return print(message)
        await self.send_notification(ctx=self._notification_channel,
                                     text=message,
                                     as_reply=False)

    def emoji_by_name(self,
                      _name: str):
        return next((x for x in self.guild.emojis if x.name == _name), None)

    async def get_help_cmds_by_user(self,
                                    ctx: discord.ext.commands.Context) -> [disco_commands.command]:
        return self.commands

    async def send_notification(self,
                                ctx: discord.ext.commands.Context | discord.abc.GuildChannel,
                                text: str,
                                as_reply: bool = False) -> None:
        """ Helper function to send notifications as required to a specific context, also as reply if required\n
        **param ctx**: context to send notification to\n
        **param text**: text of the notification body\n
        **as_reply**: send the notification as a discord reply to the context\n
        **returns**: None\n
        """

        embed = (self.default_embed('**Notification**')
                 .add_field(name='Message', value=text, inline=True)
                 .set_footer(text=f'Generated: {datetime.datetime.now()}'))
        if self._server_icon:
            embed.set_thumbnail(url=self.get_emoji(self._server_icon).url)
        await ctx.reply(embed=embed) if as_reply and (ctx.author is not None) else await ctx.send(embed=embed)

    def info_embed(self) -> discord.Embed:
        """ helper to get information embed\n
            **returns**: discord.Embed with bot info attached\n
        """
        embed = discord.Embed(color=self.default_embed_color, title='**Bot Info**\n\n',
                              description='For help, type `ub.help`\n\n')
        embed.add_field(name='Version', value=f"`{self.version}`", inline=True)
        embed.add_field(name='Boot Time', value=f"`{self._start_time}`", inline=False)
        embed.add_field(name='Current Tick', value=f"`{self._periodic_task.ticks}`", inline=True)
        embed.add_field(name='Last Tick Time', value=f"`{self._last_time}`", inline=False)
        embed.add_field(name='Cycle Time', value=f"`{self.cycle_time}` seconds", inline=True)
        if self._server_icon:
            embed.set_thumbnail(url=self.get_emoji(self._server_icon).url)
        return embed

    async def on_command_error(self,
                               ctx: discord.ext.commands.Context,
                               error) -> None:
        """ Override of discord.Bot on_command_error
            If CommandNotFound, simply reply to the user of the error.\n
            If not, raise the error naturally\n
            ***param ctx***: context to send the error message to\n
            ***param error***: error that occurred\n
            ***returns***: None\n
            **raises**: supplied error if not **discord.ext.commands.errors.CommandNotFound**
            """
        if isinstance(error, discord.ext.commands.errors.CommandNotFound):
            await ctx.reply("That command wasn't found! Type 'ub.help' for a list of all available commands.")
            return
        if isinstance(error, discord.ext.commands.errors.MissingRequiredArgument):
            await ctx.reply(f'You must fill in additional arguments for this command!\n{error}')
            return
        elif isinstance(error, discord.HTTPException):
            await ctx.reply('We are being rate limited... Please wait a few moments before trying that again.')
            return
        elif isinstance(error, InsufficientPrivilege):
            await ctx.reply(error.__str__())
            return
        elif isinstance(error, IllegalChannel):
            await ctx.reply(error.__str__())
            return
        elif isinstance(error, BotNotLoaded):
            await ctx.reply(error.__str__())
            return
        else:
            await self.__err__(f'We have encountered the following error:\n{error.__str__()}')
            raise error

    async def on_ready(self,
                       suppress_task=False) -> None:
        """ Method that is called by discord when the bot connects to the supplied guild\n
        **param suppress_task**: if True, do NOT run periodic task at the end of this call\n
        **returns**: None\n
        """
        if self._initialized:
            return

        guild_token = os.getenv('GUILD')
        handler_token = os.getenv('HANDLER')
        admin_channel_token = os.getenv('ADMIN_CHANNEL')
        notification_channel_token = os.getenv('NOTIFICATION_CHANNEL')
        admin_commands_channel_token = os.getenv('ADMIN_CMDS_CHANNEL')
        public_commands_channel_token = os.getenv('PUBLIC_CMDS_CHANNEL')

        if not guild_token:
            raise ValueError('guild token must be supplied in .env file!')
        self._guild = next((x for x in self.guilds if x.id.__str__() == guild_token), None)
        if self._guild is None:
            raise ValueError('Supplied guild is not available to this bot.\n'
                             'Please invite the bot to the guild you wish to connect to')

        self._handler = next((x for x in self._guild.members if x.id.__str__() == handler_token), None)
        self._admin_channel = get_channel_by_id(admin_channel_token,
                                                self._guild) if admin_channel_token else None
        self._notification_channel = get_channel_by_id(notification_channel_token,
                                                       self._guild) if notification_channel_token else None
        self._admin_commands_channel = get_channel_by_id(admin_commands_channel_token,
                                                         self.guild) if admin_commands_channel_token else None
        self._public_commands_channel = get_channel_by_id(public_commands_channel_token,
                                                          self.guild) if public_commands_channel_token else None

        self._initialized = True

        print(f'POST -> {datetime.datetime.now()}')

        if not suppress_task:
            await self.__begin_task__()

    async def on_task(self) -> None:
        """ callback method for periodic task to call during it's interval\n
            override this as needed
        **returns**: None\n
        """
        pass


if __name__ == '__main__':
    """ common call for .py file.
        if this file is called, the code below is run
        if this file is imported, it will not run
        """
    """ load dotenv before generating bot
    """
    dotenv.load_dotenv()

    """ generate intents for bot
        the compiler doesn't like that it thinks these properties are read-only, but they're not.
        the following boolean assignments are required for this bot to function
        """
    intents = discord.Intents(8)
    # noinspection PyDunderSlots
    intents.guilds = True
    # noinspection PyDunderSlots
    intents.members = True
    # noinspection PyDunderSlots
    intents.message_content = True
    # noinspection PyDunderSlots
    intents.messages = True

    """ generate bot
        load bot params from .env file
        do not modify this code!
        modify the .env!
        include any task args at the end!
    """
    bot = Bot('ub.',
              intents,
              command_cogs=[Commands])
    """ Start 
    """
    bot.run(os.getenv('DISCORD_TOKEN'))



#!/usr/bin/env python
""" Channels functions
# Author: irox_rl
# Purpose: General Functions of discord channels
# Version 1.0.5
#
# v1.0.5 - update clear_channel_messages to use err function for reporting
"""

# local imports #
from .err import err

# non-local imports
import discord


async def clear_channel_messages(channel: discord.channel,
                                 count: int) -> bool:
    """ Helper function to clear a channel of messages\n
        Max channel message delete count is 100\n
        **param channel**: the channel to delete messages from\n
        **param count**: number of messages to delete
        """
    """ If count is out of bounds, return 
    """
    if count > 100 | count <= 0:
        raise ValueError(f'Cannot delete that many messages!: {count}')
    """ Attempt to delete message from discord server channel specified
        Catch exceptions specific by discord API method and return
    """
    try:
        await channel.delete_messages([message async for message in channel.history(limit=count)])
        return True
    except (discord.ClientException, discord.Forbidden, discord.HTTPException) as e:
        await err(e)
        return False


def get_channel_by_id(channel_id: str,
                      guild: discord.Guild) -> discord.abc.GuildChannel | None:
    """ Get specific Guild Channel by passed ID """
    return next((x for x in guild.channels if x.id.__str__() == channel_id), None)


async def get_channel_message_by_id(channel: discord.TextChannel | discord.abc.GuildChannel,
                                    message_id: str) -> discord.Message | None:
    """ Get a specific message in a specified channel by passed ID """
    return next((x for x in [message async for message in channel.history(limit=200)] if x.id.__str__() == str(message_id)),
                None)


async def post_image(channel: discord.channel,
                     picture_file: str):
    """ Post an image to a specified channel """
    with open(picture_file, 'rb') as f:
        # noinspection PyTypeChecker
        picture = discord.File(f)
        await channel.send(file=picture)



#!/usr/bin/env python
""" Minor League E-Sports Bot Commands
# Author: irox_rl
# Purpose: General Functions and Commands
# Version 1.0.5
#
# v1.0.5 - remove err string (unused anyways)
#           hide a few commands
#           update help function for pagination
#           add get_admin_channel and get_public_channel for command predicates
"""

# local imports #
from .pagination import Pagination

# non-local imports
import datetime
import discord
from discord.ext import commands
import time


class Commands(commands.Cog):
    def __init__(self,
                 master_bot):
        self.bot = master_bot

    async def get_admin_cmds_channel(self):
        return self.bot.admin_commands_channel

    async def get_public_cmds_channel(self):
        return self.bot.public_commands_channel

    def bot_loaded(self) -> bool:
        return self.bot.loaded

    @commands.command(name='echo',
                      description='echo',
                      hidden=True)
    async def echo(self,
                   ctx: discord.ext.commands.Context,
                   *message: str):
        await ctx.send(' '.join(message))

    @commands.command(name='datetounix',
                      description='convert date to unix code for discord purposes.\n'
                                  'You must give a date in the following format: y/m/d H:M:S')
    async def datetounix(self,
                         ctx: discord.ext.commands.Context,
                         *date_string: str):
        try:
            _date = ' '.join(date_string)
            d = datetime.datetime.strptime(_date, '%y/%m/%d %H:%M:%S')
            unix_time = time.mktime(d.timetuple())
            await ctx.send(f'template: <`t:{str(int(unix_time))}:F`>  (remember to add < and > around the template!)')
            await ctx.send(f'<t:{int(unix_time)}:F>')
        except ValueError:
            await ctx.reply('You must give a date in the following format:\n'
                            '%y/%m/%d %H:%M:%S\n'
                            'do not include am/pm. use 24 hour clock.')

    @commands.command(name='help',
                      description="Show all available commands for this bot.")
    async def help(self,
                   ctx: discord.ext.commands.Context):
        cmds = await self.bot.get_help_cmds_by_user(ctx)
        sorted_commands = sorted([command for command in cmds if not command.hidden], key=lambda x: x.name)

        async def get_page(page: int,
                           as_timout: bool=False):
            emb: discord.Embed = self.bot.default_embed(f'**Commands List**\n\n',
                                                        f'Available commands for {ctx.author.mention}')
            if self.bot.server_icon:
                emb.set_thumbnail(url=self.bot.get_emoji(self.bot.server_icon).url)

            if as_timout:
                emb.add_field(name=f'**`Timeout`**',
                              value='This command has timed out. Type `[ub.help]` for help.')
                emb.set_footer(text=f'Page 1 of 1')
                return emb, 0

            elements_per_page = 5
            offset = (page - 1) * elements_per_page
            for cmd in sorted_commands[offset:offset + elements_per_page]:
                emb.add_field(name=f'**`ub.{cmd} {"  ".join([f"[{param}]" for param in cmd.clean_params])}`**',
                              value=f'{cmd.description}',
                              inline=False)
            total_pages = Pagination.compute_total_pages(len(sorted_commands),
                                                         elements_per_page)

            emb.set_footer(text=f'Page {page} of {total_pages}')
            return emb, total_pages

        await Pagination(ctx, get_page).navigate()

    @commands.command(name="info",
                      description="Get build info.")
    async def info(self,
                   ctx: discord.ext.commands.Context):
        return await ctx.reply(embed=self.bot.info_embed())

    @commands.command(name='test',
                      description='developer test function',
                      hidden=True)
    async def test(self,
                   ctx: discord.ext.commands.Context):
        await ctx.send('guh? huh? who what? where am i?')

    @commands.Cog.listener()
    async def on_message(self,
                         message: discord.Message):
        if message.author == self.bot.user:
            return

        if 'nice' == message.content.lower():
            await message.reply('nice')



#!/usr/bin/env python
""" Pagination handling for discord API
# Author: irox_rl
# Purpose: Handle pagination efforts for embeds
# Version 1.0.5
#
# v1.0.5 - initial release, developed on ub.help command
"""

# local imports #
from typing import Callable, Optional

# non-local imports
import discord
from discord.ext import commands


class Pagination(discord.ui.View):
    def __init__(self,
                 ctx: discord.ext.commands.Context,
                 get_page: Callable):
        self.ctx = ctx
        self.msg: discord.Message | None = None
        self.get_page = get_page
        self.total_pages: Optional[int] = None
        self.index = 1
        super().__init__(timeout=100)

    async def interaction_check(self,
                                interaction: discord.Interaction) -> bool:
        return True

    async def navigate(self):
        embed, self.total_pages = await self.get_page(self.index)
        if self.total_pages == 1:
            if not self.msg:
                self.msg = await self.ctx.send(embed=embed)
            else:
                await self.msg.edit(embed=embed)
        elif self.total_pages > 1:
            self.update_buttons()
            if not self.msg:
                self.msg = await self.ctx.send(embed=embed, view=self)
            else:
                await self.msg.edit(embed=embed, view=self)

    async def edit_page(self,
                        interaction: discord.Interaction):
        embed, self.total_pages = await self.get_page(self.index)
        self.update_buttons()
        if not self.msg:
            self.msg = await self.ctx.send(embed=embed, view=self)
        else:
            await self.msg.edit(embed=embed, view=self)

        # this WILL throw. but it's fine. we just need it for the interaction complete
        # but we don't want to be bombarded by messages
        try:
            await interaction.response.send_message('', ephemeral=True)
        except discord.HTTPException:
            return

    def update_buttons(self):
        self.children[0].disabled = self.index == 1
        self.children[1].disabled = self.index == self.total_pages

    @discord.ui.button(emoji="◀️", style=discord.ButtonStyle.gray)
    async def previous(self, interaction: discord.Interaction, button: discord.Button):
        self.index -= 1
        await self.edit_page(interaction)

    @discord.ui.button(emoji="▶️", style=discord.ButtonStyle.gray)
    async def next(self, interaction: discord.Interaction, button: discord.Button):
        self.index += 1
        await self.edit_page(interaction)

    async def on_timeout(self) -> None:
        embed, self.total_pages = await self.get_page(0,
                                                      as_timout=True)
        if not self.msg:
            self.msg = await self.ctx.send(embed=embed)
        else:
            await self.msg.edit(embed=embed, view=None)

    @staticmethod
    def compute_total_pages(total_results: int, results_per_page: int) -> int:
        return ((total_results - 1) // results_per_page) + 1



#!/usr/bin/env python
""" error
# Author: irox_rl
# Purpose: show errors? also register callback for error messages
# Version 1.0.5
#
#       v1.0.5 - include BotNotLoaded err for predicate checks on commands
"""

# local imports #

# non-local imports
from discord.ext import commands

from typing import Callable

err_callback: Callable = None


class BotNotLoaded(commands.CheckFailure):
    pass


class InsufficientPrivilege(commands.CheckFailure):
    pass


class IllegalChannel(commands.CheckFailure):
    pass


async def err(message: str | Exception) -> None:
    """ Helper function to send error or notification messages to notify channel with a single parameter.\n
        **If a notification channel does not exist**, the notification is printed to console instead\n
        **param message**: message to report\n
        **returns**: None\n
        """
    global err_callback
    if not message:
        return
    if not err_callback:
        return print(message)
    await err_callback(message)


def register_callback(callback: Callable):
    global err_callback
    err_callback = callback
